// obj.hxx -- routines to handle loading scenery and building the plib
//            scene graph.
//
// Written by Curtis Olson, started October 1997.
//
// Copyright (C) 1997  Curtis L. Olson  - http://www.flightgear.org/~curt
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//
// $Id$


#ifndef _SG_OBJ_HXX
#define _SG_OBJ_HXX

#ifndef __cplusplus
# error This library requires C++
#endif

#include <boost/lexical_cast.hpp>

#include <simgear/compiler.h>

#include <string>

#include <osg/Node>
#include <osg/Group>

using std::string;

using boost::lexical_cast;
using boost::bad_lexical_cast;

class SGMaterialLib;
namespace simgear {
class SGReaderWriterOptions;
}

osg::Node*
SGLoadBTG(const std::string& path, 
          const simgear::SGReaderWriterOptions* options);

#endif // _SG_OBJ_HXX
