
#define HAVE_SYS_TIME_H
/* #undef HAVE_SYS_TIMEB_H */
#define HAVE_UNISTD_H 1


#define HAVE_GETTIMEOFDAY
/* #undef HAVE_GETLOCALTIME */
/* #undef HAVE_FTIME */
/* #undef HAVE_RINT */
#define HAVE_TIMEGM

/* #undef HAVE_STD_ISNAN */
#define HAVE_WORKING_STD_REGEX
/* #undef HAVE_WINDOWS_H */
#define HAVE_MKDTEMP
/* #undef HAVE_AL_EXT_H */
#define HAVE_STD_INDEX_SEQUENCE
#define HAVE_STD_REMOVE_CV_T
/* #undef HAVE_STD_REMOVE_CVREF_T */
#define HAVE_STD_ENABLE_IF_T
#define HAVE_STD_BOOL_CONSTANT

/* #undef GCC_ATOMIC_BUILTINS_FOUND */

#define OPENVG
#define SYSTEM_EXPAT
#define ENABLE_SOUND
/* #undef USE_AEONWAVE */
/* #undef SG_SOUND_USES_OPENALSOFT */
/* #undef ENABLE_SIMD */
/* #undef ENABLE_GDAL */

#define SIMGEAR_USING_CARES_DNS
#define SG_FFMPEG
