/*
 * Copyright (C) 2006-2007 Tim Moore timoore@redhat.com
 * Copyright (C) 2008 Till Busch buti@bux.at
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 *
 */
#ifndef SGREADERWRITERXML_HXX
#define SGREADERWRITERXML_HXX 1
#include <osgDB/Registry>
namespace simgear
{

class SGReaderWriterXML : public osgDB::ReaderWriter
{
public:
    SGReaderWriterXML();
    virtual ~SGReaderWriterXML();

    virtual const char* className() const;

    virtual ReadResult readNode(const std::string& fileName,
                                const osgDB::Options* options)
    const;
};

}
#endif

