////
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the
// Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
// Boston, MA  02110-1301, USA.

#pragma once

#include <osg/Program>

#include <simgear/misc/sg_path.hxx>

/**
 * @brief wrapper around osg::Program to allow detecting shader/link
 * errors in the GLSL code at runtime, and reporting of them
 * 
 */
class SGProgram : public osg::Program
{
public:
    SGProgram();

    SGProgram(const SGProgram& rhs, const osg::CopyOp& copyop = osg::CopyOp::SHALLOW_COPY);

    META_StateAttribute(simgear, SGProgram, PROGRAM);

    void apply(osg::State& state) const override;

    enum ErrorCheckState {
        NotApplied,
        AppliedOk,
        FailedToApply
    };

    void setEffectFilePath(const SGPath& p);

private:
    SGPath _effectFilePath;
    mutable ErrorCheckState _checkState = NotApplied;
};
